# -*- coding: utf-8 -*-

"""
Shared structural classes
"""

import time
import csv
from typing import Any, Iterator, Tuple
from lbsnstructure import lbsnstructure_pb2 as lbsn


class Coordinates():
    """Lat lng coordinates in Decimal Degrees WGS1984"""

    def __init__(self, lat=None, lng=None):
        if lat is None:
            lat = 0
        if lng is None:
            lng = 0
        self.lat: float = self.lat_within_range(lat)
        self.lng: float = self.lng_within_range(lng)

    @classmethod
    def lat_within_range(cls, lat):
        if not -90 <= lat <= 90:
            raise ValueError('Latitude outside allowed range')
        return lat

    @classmethod
    def lng_within_range(cls, lng):
        if not -180 <= lng <= 180:
            raise ValueError('Longitude outside allowed range')
        return lng


class LBSNRecordDicts():
    """Parent structure to store collected lbsnstructures"""

    def __init__(self):
        self.lbsn_origin_dict = dict()
        self.lbsn_country_dict = dict()
        self.lbsn_city_dict = dict()
        self.lbsn_place_dict = dict()
        self.lbsn_user_group_dict = dict()
        self.lbsn_user_dict = dict()
        self.lbsn_post_dict = dict()
        self.lbsn_post_reaction_dict = dict()
        self.lbsn_relationship_dict = dict()
        self.key_hashes = {lbsn.Origin.DESCRIPTOR.name: set(),
                           lbsn.Post.DESCRIPTOR.name: set(),
                           lbsn.Country.DESCRIPTOR.name: set(),
                           lbsn.City.DESCRIPTOR.name: set(),
                           lbsn.Place.DESCRIPTOR.name: set(),
                           lbsn.UserGroup.DESCRIPTOR.name: set(),
                           lbsn.User.DESCRIPTOR.name: set(),
                           lbsn.PostReaction.DESCRIPTOR.name: set(),
                           lbsn.Relationship.DESCRIPTOR.name: set()}
        self.count_glob = 0  # total number of records added
        self.count_glob_total = 0
        self.count_dup_merge = 0  # number of duplicate records merged
        self.count_dup_merge_total = 0
        # returns all recordsDicts in correct order,
        # with names as references (tuple)
        self.all_dicts = [
            (self.lbsn_origin_dict, lbsn.Origin().DESCRIPTOR.name),
            (self.lbsn_country_dict, lbsn.Country().DESCRIPTOR.name),
            (self.lbsn_city_dict, lbsn.City().DESCRIPTOR.name),
            (self.lbsn_place_dict, lbsn.Place().DESCRIPTOR.name),
            (self.lbsn_user_group_dict, lbsn.UserGroup().DESCRIPTOR.name),
            (self.lbsn_user_dict, lbsn.User().DESCRIPTOR.name),
            (self.lbsn_post_dict, lbsn.Post().DESCRIPTOR.name),
            (self.lbsn_post_reaction_dict, lbsn.PostReaction().DESCRIPTOR.name),
            (self.lbsn_relationship_dict, lbsn.Relationship().DESCRIPTOR.name)
        ]

    def get_current_count(self):
        """Return count of collected records"""
        count_glob = self.count_glob
        return count_glob

    def get_all_records(self) -> Iterator[Tuple[Any, str]]:
        """Returns tuple of 1) all records from self
        in correct order using all_dicts and 2) Type of record

        Order is: lbsn.Country(), lbsn.City(), lbsn.Place(), lbsn.UserGroup(),
        lbsn.User(), lbsn.Post(), lbsn.PostReaction(), lbsn.Relationship()
        """
        for records_dict in self.all_dicts:
            type_name = records_dict[1]
            for record in records_dict[0].values():
                yield record, type_name

    def get_type_counts(self):
        """Get record counts per lbsn type"""
        count_list = []
        for x, y in self.key_hashes.items():
            count_list.append(f'{x}: {len(y)} ')
        return ''.join(count_list)

    def update_key_hash(self, record, key=None):
        """Update key-hash with record

        Keep lists of pkeys for each type
        this can be used to check for duplicates or to get a
        total count for each type of records (Number of unique
        Users, Countries, Places etc.)
        in this case we assume that origin_id remains the same
        in each program iteration!
        """
        if key is not None:
            self.key_hashes[record.DESCRIPTOR.name].add(key)
            return
        if record.DESCRIPTOR.name == lbsn.Relationship().DESCRIPTOR.name:
            # we need the complete uuid of both entities for
            # relationships because they can span different origin_ids
            self.key_hashes[record.DESCRIPTOR.name].add(
                f'{record.pkey.relation_to.origin.origin_id}'
                f'{record.pkey.relation_to.id}'
                f'{record.pkey.relation_from.origin.origin_id}'
                f'{record.pkey.relation_from.id}'
                f'{record.relationship_type}')
        else:
            # all other entities can be globally uniquely
            # identified by their local guid
            self.key_hashes[record.DESCRIPTOR.name].add(record.pkey.id)

    def clear(self):
        """Clears all records from all dicts
        """
        for lbsn_dict, __ in self.all_dicts:
            lbsn_dict.clear()
        self.count_glob_total += self.count_glob
        self.count_glob = 0
        self.count_dup_merge_total += self.count_dup_merge
        self.count_dup_merge = 0

    def deep_compare_merge_messages(self, old_record, new_record):
        """Do a deep compare & merge of two lbsn records

        ProtoBuf MergeFrom does a fine job
        only problem is it concatenates repeate strings,
        which may result in duplicate entries
        we take care of this prior to submission (see submitData classes)
        """
        old_record.MergeFrom(new_record)
        # for descriptor in oldRecord.DESCRIPTOR.fields:
        #        if descriptor.label == descriptor.LABEL_REPEATED:
        #            if value_old == value_new:
        #                return oldRecord
        #            elif not value_old:
        #                newEntries = value_new
        #            else:
        #                # only add difference (e.g. = new values)
        #                newEntries = list(set(value_new) - set(value_old))
        #            x = getattr(oldRecord, descriptor.name)
        #            x.extend(newEntries)
        return old_record

    def add_records_to_dict(self, records):
        """Add single or multiple lbsnrecords to dicts"""
        if not records:
            return
        if isinstance(records, (list,)):
            for record in records:
                self.add_record_to_dict(record)
        else:
            record = records
            self.add_record_to_dict(record)

    def dict_selector(self, record):
        """ Get dictionary by record type name"""
        dict_switcher = {
            lbsn.Post().DESCRIPTOR.name: self.lbsn_post_dict,
            lbsn.Country().DESCRIPTOR.name: self.lbsn_country_dict,
            lbsn.City().DESCRIPTOR.name: self.lbsn_city_dict,
            lbsn.Place().DESCRIPTOR.name: self.lbsn_place_dict,
            lbsn.PostReaction().DESCRIPTOR.name: self.lbsn_post_reaction_dict,
            lbsn.User().DESCRIPTOR.name: self.lbsn_user_dict,
            lbsn.UserGroup().DESCRIPTOR.name: self.lbsn_user_group_dict,
            lbsn.Origin().DESCRIPTOR.name: self.lbsn_origin_dict
        }
        return dict_switcher.get(record.DESCRIPTOR.name)

    def add_record_to_dict(self, newrecord):
        """Add single lbsn record to dict"""
        sel_dict = self.dict_selector(newrecord)
        try:
            pkey_id = newrecord.pkey.id
        except AttributeError:
            # inexpensive try that is raised rarely
            # lbsn origins have no composite pkey
            # use origin_id instead
            pkey_id = newrecord.origin_id
        if pkey_id in sel_dict:
            oldrecord = sel_dict[pkey_id]
            # oldrecord will be modified/updated
            self.merge_existing_records(oldrecord, newrecord)
            self.count_dup_merge += 1
        else:
            # just count new entries
            self.count_glob += 1
            # update keyHash only necessary for new record
            self.update_key_hash(newrecord, pkey_id)
            sel_dict[pkey_id] = newrecord

    def add_relationship_to_dict(self, newrelationship):
        """Add single lbsn relationship to dict"""
        pkey_id = f'{newrelationship.pkey.relation_to.origin.origin_id}' \
                  f'{newrelationship.pkey.relation_to.id}' \
                  f'{newrelationship.pkey.relation_from.origin.origin_id}' \
                  f'{newrelationship.pkey.relation_from.id}' \
                  f'{newrelationship.relationship_type}'
        if pkey_id not in self.lbsn_relationship_dict:
            # self.count_progress_report()
            self.lbsn_relationship_dict[pkey_id] = newrelationship
            # update keyHash only necessary for new record
            self.update_key_hash(newrelationship)

    @classmethod
    def merge_existing_records(cls, oldrecord, newrecord):
        """Merge two lbsn records

        Basic Compare function for GUIDS
        First check if length of both ProtoBuf Messages are the same
        """
        old_record_string = oldrecord.SerializeToString()
        new_record_string = newrecord.SerializeToString()
        if not len(old_record_string) == len(new_record_string):
            # no need to do anything if same lengt
            oldrecord.MergeFrom(newrecord)
            # updatedrecord = self.deepCompareMergeMessages(oldrecord,
            #                                               newrecord)


class GeocodeLocations():
    """Class for geocoding of text to lat/lng values"""

    def __init__(self):
        self.geocode_dict = dict()

    def load_geocodelist(self, file):
        # read each unsorted file and sort lines based on datetime (as string)
        with open(file, newline='', encoding='utf8') as fhandle:
            # next(f) #Skip Headerrow
            locationfile_list = csv.reader(
                fhandle, delimiter=',', quotechar='', quoting=csv.QUOTE_NONE)
            for location_geocode in locationfile_list:
                self.geocode_dict[location_geocode[2].replace(
                    ';', ',')] = (float(location_geocode[0]),  # lat
                                  location_geocode[1])  # lng
        print(f'Loaded {len(self.geocode_dict)} geocodes.')


class TimeMonitor():
    """Utility to report processing speed

    Once initialized, the start time will be
    recorded and can be stopped at any time with
    stop_time(), which will return the time passed
    in a text readable time format.
    """

    def __init__(self):
        self.now = time.time()

    def stop_time(self):
        """Returns a text with time passed since self.now"""
        later = time.time()
        hours, rem = divmod(later-self.now, 3600)
        minutes, seconds = divmod(rem, 60)
        # difference = int(later - self.now)
        report_msg = f'{int(hours):0>2} Hours {int(minutes):0>2} ' \
                     f'Minutes and {seconds:05.2f} Seconds passed.'
        return report_msg



"""Script to allow argdown parse_args to markdown conversion"""

import inspect
from lbsntransform import BaseConfig
from lbsntransform import __version__


def extract_argscode():
    """Extracts command line args code to separate file

    Preparation step for processing with argdown
    """
    # open file to output source code
    source_file = open("parse_args.py", "w")
    # extract source code of parse_args
    parse_args_source = inspect.getsource(BaseConfig.parse_args)
    # remove first line
    parse_args_source = parse_args_source[parse_args_source.index('\n')+1:]
    # unindent all other lines
    parse_args_source = parse_args_source.lstrip().replace('\n        ', '\n')
    # replace version string
    parse_args_source = parse_args_source.replace(
        'lbsntransform {__version__}', f'lbsntransform {__version__}')
    # replace package name
    parse_args_source = parse_args_source.replace(
        'usage: argdown', 'usage: lbsntransform')
    # write argdown and argparse imports first
    source_file.write('import argparse\n')
    source_file.write('import argdown\n')
    source_file.write('from pathlib import Path\n')
    # fix argparse name
    parse_args_source = parse_args_source.replace(
        'ArgumentParser()', 'ArgumentParser(prog="lbsntransform")')
    # replace two empty spaces on str line ends with non-breaking spaces,
    # protecting line breaks during python-markdown conversion
    parse_args_source = parse_args_source.replace(
        "'  '\n", "'&nbsp;&nbsp;<br>'\n")
    parse_args_source = parse_args_source.replace(
        "  '\n", "&nbsp;&nbsp;<br>'\n")
    # protected indent
    parse_args_source = parse_args_source.replace(
        "'    ", "'&nbsp;&nbsp;&nbsp;&nbsp;")
    # write prepared source code
    source_file.write(parse_args_source)
    source_file.close()


# run script
extract_argscode()



# -*- coding: utf-8 -*-

"""
Config module for parsing input args for lbsntransform package.
"""

# pylint: disable=no-member

import argparse
import logging
from pathlib import Path
from typing import List, Tuple

from shapely import geos

from lbsnstructure import lbsnstructure_pb2 as lbsn
from lbsntransform import __version__


class BaseConfig():
    """Base config class for handling typical input params"""

    def __init__(self):
        """Set Default Config options here

        - or define options as input args
        """
        self.origin = 0
        self.is_local_input = False
        self.local_file_type = 'json'
        self.input_path = None
        self.is_stacked_json = False
        self.is_line_separated_json = False
        self.dbuser_input = 'example-user-name'
        self.dbpassword_input = 'example-user-password'
        self.dbserveraddress_input = '222.22.222.22'
        self.dbserverport_input = 5432
        self.dbformat_input = 'json'
        self.dbname_input = 'test_db2'
        self.dbname_hllworker = None
        self.dbuser_hllworker = None
        self.dbpassword_hllworker = None
        self.dbserveraddress_hllworker = None
        self.dbserverport_hllworker = 5432
        self.dbname_hllworker = None
        self.dbuser_output = None
        self.dbformat_output = 'lbsn'
        self.dbpassword_output = None
        self.dbserveraddress_output = None
        self.dbserverport_output = 5432
        self.dbname_output = None
        self.transferlimit = None
        self.transfer_count = 50000
        self.number_of_records_to_fetch = 10000
        self.transfer_reactions = True
        self.disable_reactionpost_ref = False
        self.ignore_non_geotagged = False
        self.startwith_db_rownumber = None
        self.endwith_db_rownumber = None
        self.debug_mode = 'INFO'
        self.geocode_locations = False
        self.ignore_input_source_list = False
        self.input_lbsn_type = None
        self.map_relations = False
        self.csv_output = False
        self.csv_suppress_linebreaks = True
        self.csv_delim = None
        self.use_csv_dictreader = None
        self.recursive_load = False
        self.skip_until_file = None
        self.min_geoaccuracy = None
        self.logging_level = logging.INFO
        self.source_web = False
        self.skip_until_record = None
        self.zip_records = None
        self.include_lbsn_objects = []
        self.include_lbsn_bases = None
        self.override_lbsn_query_schema = None
        self.mappings_path = None
        self.dry_run = None
        self.hmac_key = None

        BaseConfig.set_options()

    def parse_args(self):
        """Process input *args

        All args are optional, but some groups need to be defined together.
        """
        parser = argparse.ArgumentParser()
        parser.add_argument('--version',
                            action='version',
                            version=f'lbsntransform {__version__}')
        parser.add_argument('-o', "--origin",
                            default=0,
                            help='Input source type (id). '
                            '  '
                            '  '
                            '* Defaults to `0`: LBSN  '
                            '  '
                            'Other possible values:  '
                            '* `1` - Instagram  '
                            '* `2` - Flickr  '
                            '* `21` - Flickr YFCC100M  '
                            '* `3` - Twitter  ',
                            type=int)
        parser.add_argument("--dry-run",
                            action='store_true',
                            help='Perform a trial run '
                            '  '
                            'with no changes made '
                            'to database/output')
        # Local Input
        local_input_args = parser.add_argument_group('Local Input')
        local_input_args.add_argument('-l', "--file_input",
                                      action='store_true',
                                      help='This flag enables file input  '
                                      '  '
                                      '(instead of reading data from a database). '
                                      '  '
                                      '  '
                                      '* To specify which files to process, see '
                                      'parameter `--input_path_url`.  '
                                      '* To specify file types, e.g. whether to '
                                      'process data from `json` '
                                      'or `csv`, or from URLs,  '
                                      '  see `--file_type`  ')
        local_input_args.add_argument("--file_type",
                                      default='json',
                                      help='Specify filetype  '
                                      '  '
                                      ' (`json`, `csv` etc.) '
                                      '  '
                                      '  '
                                      '* only applies if `--file_input` is used.  ',
                                      type=str)
        local_input_args.add_argument("--input_path_url",
                                      default="01_Input",
                                      help='Path to input folder.  '
                                      '  '
                                      '* If not provided, subfolder `./01_Input/` '
                                      ' will be used.  '
                                      '* You can also provide a web-url, '
                                      'starting with `http(s)`  '
                                      '* URLs will be accessed using '
                                      '`requests.get(url, stream=True)`.  '
                                      '* To separate multiple urls, use '
                                      'semicolon (`;`). In this case, see also '
                                      '`--zip_records`.  ',
                                      type=str)
        local_input_args.add_argument("--is_stacked_json",
                                      action='store_true',
                                      default=False,
                                      help='Input is stacked json. '
                                      '  '
                                      '  '
                                      '* The typical form of json is '
                                      '`[{json1},{json2}]`  '
                                      '* If `--is_stacked_json` is set, '
                                      'it will process stacked jsons in the form '
                                      'of `{json1}{json2}` (no comma)  ')
        local_input_args.add_argument("--is_line_separated_json",
                                      action='store_true',
                                      default=False,
                                      help='Json is line separated '
                                      '  '
                                      '  '
                                      '* The typical form is '
                                      '`[{json1},{json2}]`  '
                                      '* If `--is_line_separated_json` is set, '
                                      'it will process stacked jsons in the form '
                                      'of `{json1}\n{json2}` (with linebreak)  '
                                      '* Unix style linebreaks (CR) will be used '
                                      'across platforms  '
                                      '* Windows users, use (e.g.) notepad++ to '
                                      'convert from Windows style linebreaks (CRLF)  ')
        # HLL Worker
        hllworker_args = parser.add_argument_group('HLL Worker')
        hllworker_args.add_argument("--dbpassword_hllworker",
                                    help='Password for hllworker db  '
                                    '  '
                                    '  '
                                    '* If reading data into `hlldb`, all '
                                    'HLL Worker parameters must be supplied by'
                                    'default.  '
                                    '* You can substitute hlldb parameters here  '
                                    '* In this case, lbsntransform will use '
                                    'hlldb to convert and union hll sets '
                                    '_and_ to store output results  '
                                    '* Currently, this re-use of hlldb requires '
                                    'to supply the same set of parameters twice  '
                                    '* For separation of concerns, it is recommended '
                                    'to use a separate HLL Worker database  ',
                                    type=str)
        hllworker_args.add_argument("--dbuser_hllworker",
                                    default="postgres",
                                    help='Username for hllworker db.',
                                    type=str)
        hllworker_args.add_argument("--dbserveraddress_hllworker",
                                    help='IP for hllworker db  '
                                    '  '
                                    '  '
                                    '* e.g. `111.11.11.11`  '
                                    '* Optionally add '
                                    'port the to use, e.g. `111.11.11.11:5432`.  '
                                    '* `5432` is the default port  ',
                                    type=str)
        hllworker_args.add_argument("--dbname_hllworker",
                                    help='DB name for hllworker db '
                                    '  '
                                    '  '
                                    '* e.g. `hllworkerdb`  ',
                                    type=str)
        # DB Output
        dboutput_args = parser.add_argument_group('DB Output')
        dboutput_args.add_argument('-p', "--dbpassword_output",
                                   help='Password for out-db  '
                                   '  '
                                   '(postgres raw/hll db)',
                                   type=str)
        dboutput_args.add_argument('-u', "--dbuser_output",
                                   default="postgres",
                                   help='Username for out-db.  '
                                   '  '
                                   'Default: `example-user-name2`',
                                   type=str)
        dboutput_args.add_argument('-a', "--dbserveraddress_output",
                                   help='IP for output db, '
                                   '  '
                                   '  '
                                   '* e.g. `111.11.11.11`  '
                                   '* Optionally add '
                                   'port to use, e.g. `111.11.11.11:5432`.  '
                                   '* `5432` is the default port  ',
                                   type=str)
        dboutput_args.add_argument('-n', "--dbname_output",
                                   help='DB name for output db '
                                   '  '
                                   '  '
                                   '* e.g. `rawdb` or `hlldb`  ',
                                   type=str)
        dboutput_args.add_argument("--dbformat_output",
                                   default="lbsn",
                                   help='Format of the out-db. '
                                   '  '
                                   '  '
                                   '* Either `hll` or `lbsn`.  '
                                   '* This setting '
                                   'affects how data is stored, either '
                                   'in anonymized and aggregate form (`hll`), or in '
                                   'the [lbsn raw structure](https://lbsn.vgiscience.org/structure/) '
                                   '(`lbsn`).  ',
                                   type=str)
        # DB Input
        dbinput_args = parser.add_argument_group('DB Input')
        dbinput_args.add_argument("--dbpassword_input",
                                  help='Password for input-db',
                                  type=str)
        dbinput_args.add_argument("--dbuser_input",
                                  default="postgres",
                                  help='Username for input-db.',
                                  type=str)
        dbinput_args.add_argument("--dbserveraddress_input",
                                  help='IP for input-db, '
                                  '  '
                                  '  '
                                  '* e.g. `111.11.11.11`  '
                                  '* Optionally add port to use, e.g. '
                                  '`111.11.11.11:5432`.  '
                                  '* `5432` is the default port  ',
                                  type=str)
        dbinput_args.add_argument("--dbname_input",
                                  help='DB name for input-db, '
                                  '  '
                                  '  '
                                  '* e.g.: `rawdb`  ',
                                  type=str)
        dbinput_args.add_argument("--dbformat_input",
                                  default="json",
                                  help='Format of the input-db. '
                                  '  '
                                  '  '
                                  '* Either `lbsn` or `json`  '
                                  '* If lbsn is used, the native lbsn raw '
                                  'input mapping (`0`) will be used  '
                                  '* If `json` is used, a custom mapping for '
                                  'json must be provided, for mapping database '
                                  'json\'s to the lbsn structure. See '
                                  '[input mappings](https://lbsn.vgiscience.org/lbsntransform/docs/mappings/#input-mappings)  ',
                                  type=str)
        # Additional args
        settings_args = parser.add_argument_group('Additional settings')
        settings_args.add_argument('-t', "--transferlimit",
                                   help='Abort after x records. '
                                   '  '
                                   '  '
                                   '* This can be used to limit the number of '
                                   'records that will be processed.  '
                                   '* e.g. `--transferlimit 10000` will process '
                                   'the first 10000 input records  '
                                   '* Defaults to None (= process all)  '
                                   '* Note that one input record can map to '
                                   'many output records. This number applies to '
                                   'the number of input records, not the output count.  ',
                                   type=int)
        settings_args.add_argument("--transfer_count",
                                   default=50000,
                                   help='Transfer batch limit x. '
                                   '  '
                                   '  '
                                   '* Defines after how many '
                                   'parsed records the results will be '
                                   'transferred to the DB.  '
                                   '* Defaults to 50000  '
                                   '* If you have a slow server, but a fast machine, larger values '
                                   'improve speed because duplicate '
                                   'check happens in Python, and not in '
                                   'Postgres coalesce;  '
                                   '* However, larger values require '
                                   'more local memory. If you have a fast server, but a slow machine,  '
                                   'Try if smaller batch `--transfer_count` (e.g. 5000) improve spead.  '
                                   '  '
                                   '  '
                                   '!!! note  '
                                   '    Use `--transferlimit` to limit the '
                                   '    total number of records transferred. `--transfer_count` '
                                   '    instead defines the _batch_ count that is used to transfer '
                                   '    data incrementally.  ',
                                   type=int)
        settings_args.add_argument("--records_tofetch",
                                   default=10000,
                                   help='Fetch x records /batch. '
                                   '  '
                                   '  '
                                   '* If retrieving data from a db (`lbsn`), '
                                   'limit the number of records to fetch at once.  '
                                   '* Defaults to 10000  ',
                                   type=int)
        settings_args.add_argument(
            "--disable_transfer_reactions",
            action='store_true',
            help='Disable reactions. '
            '  '
            '  '
            '* If set, processing of lbsn reactions will be skipped,  '
            '* only original posts are transferred.  '
            '* This is usefull to reduce '
            'processing and data footprint for some service data, '
            'e.g. for Twitter, with a large '
            'number of reactions containing little original content.  ')
        settings_args.add_argument(
            "--disable_reaction_post_referencing",
            action='store_true',
            default=False,
            help='Disable reactions-refs. '
            '  '
            '  '
            'Enable this option in args '
            'to prevent empty posts being stored '
            'due to Foreign-Key-Exists Requirement. '
            '  '
            'Possible parameters:  '
            '* `0` = Save Original Tweets of Retweets as `posts`;  '
            '* `1` = do not store Original Tweets of Retweets;  '
            '* `2` = !Not implemented: Store Original Tweets of Retweets as '
            '`post_reactions`  ')
        settings_args.add_argument("--ignore_non_geotagged",
                                   action='store_true',
                                   help='Ignore none-geotagged. '
                                   '  '
                                   '  '
                                   'If set, posts that are not geotagged '
                                   'are ignored during processing.')
        settings_args.add_argument("--startwith_db_rownumber",
                                   help='Start with db row x. '
                                   '  '
                                   '  '
                                   '* Provide a number (row-id) to start '
                                   'processing from live db.  '
                                   '* If input db type '
                                   'is `lbsn`, provide the primary key '
                                   'to start from '
                                   '(e.g. post_guid, place_guid etc.).  '
                                   '* This flag will only work if processing a '
                                   'single lbsn object (e.g. lbsnPost).  ',
                                   type=str)
        settings_args.add_argument("--endwith_db_rownumber",
                                   help='End with db row x. '
                                   '  '
                                   '  '
                                   'Provide a number (row-id) to end '
                                   'processing from live db', type=int)
        settings_args.add_argument("--debug_mode",
                                   help='Enable debug mode.',
                                   type=str)
        settings_args.add_argument("--geocode_locations",
                                   help='Path to loc-geocodes. '
                                   '  '
                                   '  '
                                   '* Provide path to a CSV file with '
                                   'location geocodes  '
                                   '* CSV Header must be: '
                                   '`lat, lng, name`).  '
                                   '* This can be used in mappings to assign '
                                   'coordinates (lat, lng) '
                                   'to use provided locations as text  ',
                                   type=str)
        settings_args.add_argument("--ignore_input_source_list",
                                   help='Path to input ignore. '
                                   '  '
                                   '  '
                                   'Provide a path to a list of input_source '
                                   'types that will be ignored (e.g. to '
                                   'ignore certain bots etc.)',
                                   type=str)
        settings_args.add_argument("--mappings_path",
                                   help='Path mappings folder. '
                                   '  '
                                   '  '
                                   'Provide a path to a custom folder '
                                   'that contains '
                                   'one or more input mapping modules (`*.py`). '
                                   '  '
                                   '  '
                                   '* Have a look at the two sample mappings '
                                   'in [the resources folder](https://gitlab.vgiscience.de/lbsn/lbsntransform/-/tree/master/resources/mappings).  '
                                   '* See how to define custom input mappings '
                                   'in the [docs](https://lbsn.vgiscience.org/lbsntransform/docs/mappings/#input-mappings)  ',
                                   type=str)
        settings_args.add_argument("--input_lbsn_type",
                                   help='Input sub-type '
                                   '  '
                                   '  '
                                   '* e.g. `post`, `profile`, '
                                   '`friendslist`, `followerslist` etc.  '
                                   '* This can be used to select an appropiate '
                                   'mapping procedure in a single mapping module.  ',
                                   type=str)
        settings_args.add_argument("--map_full_relations",
                                   action='store_true',
                                   help='Map full relations. '
                                   '  '
                                   '  '
                                   'Set to true to map full relations, '
                                   'e.g. many-to-many relationships, '
                                   'such as `user_follows`, '
                                   '`user_friend`, or `user_mentions` etc. are '
                                   'mapped in a separate table. '
                                   'Defaults to False.')
        settings_args.add_argument("--csv_output",
                                   action='store_true',
                                   help='Store to local CSV. '
                                   '  '
                                   '  '
                                   'If set, will store all '
                                   'submit values to local CSV instead. '
                                   'Currently, this type of output is not available.')
        settings_args.add_argument("--csv_allow_linebreaks",
                                   action='store_true',
                                   help=repr('Disable linebreak-rem. '
                                   '  '
                                   '  '
                                   'If set, will not '
                                   'remove intext-linebreaks (`\r` or `\n`) '
                                   'in output CSVs')[1:-1])
        settings_args.add_argument("--csv_delimiter",
                                   default=",",
                                   help=repr('CSV delimiter. '
                                             '  '
                                             '  '
                                             '* Provide the CSV delimiter to be used.  '
                                             '* Default is comma (`,`).  '
                                             '* Note: to pass tab, '
                                             'use variable substitution (`$"\t"`)  ')[1:-1],
                                   type=str)
        settings_args.add_argument("--use_csv_dictreader",
                                   action='store_true',
                                   help='Use csv.DictReader. '
                                   '  '
                                   '  '
                                   'By default, CSVs will be read line by line,  '
                                   'using the standard csv.reader().  '
                                   '  '
                                   'This will enable [csv.DictReader()](https://docs.python.org/3/library/csv.html#csv.DictReader),  '
                                   'which allows to access CSV fields by name in mappings.  '
                                   '  '
                                   'A CSV with a header is required for this setting to work.  '
                                   '  '
                                   'Note that `csv.DictReader()` may be slower than the default `csv.reader()`.')
        settings_args.add_argument("--recursive_load",
                                   action='store_true', default=False,
                                   help='Recursive local sub dirs. '
                                   '  '
                                   '  '
                                   'If set, process input directories '
                                   'recursively (default depth: `2`)')
        settings_args.add_argument("--skip_until_file",
                                   help='Skip until file x. '
                                   '  '
                                   '  '
                                   'If local input, skip all files '
                                   'until file with name `x` appears '
                                   '(default: start immediately)',
                                   type=str)
        settings_args.add_argument("--skip_until_record",
                                   help='Skip until record x. '
                                   '  '
                                   '  '
                                   'If local input, skip all records '
                                   'until record `x` '
                                   '(default: start with first)',
                                   type=int)
        settings_args.add_argument("--zip_records",
                                   action='store_true', default=False,
                                   help='Zip records parallel. '
                                   '  '
                                   '  '
                                   '* Use this flag to zip records of '
                                   'multiple input files  '
                                   '* e.g. `List1[A,B,C]`, `List2[1,2,3]` will be '
                                   'combined (zipped) on read to '
                                   '`List[A1,B2,C3]`  ')
        settings_args.add_argument("--min_geoaccuracy",
                                   help='Min geoaccuracy to use '
                                   '  '
                                   '  '
                                   'Set to `latlng`, `place`, '
                                   'or `city` to limit processing of records based '
                                   'on mininum geoaccuracy (default: no limit)',
                                   type=str)
        settings_args.add_argument("--include_lbsn_objects",
                                   help='lbsn objects to process '
                                   '  '
                                   '  '
                                   'If processing from lbsn db (`rawdb`), '
                                   'provide a comma separated list '
                                   'of [lbsn objects][1] to include. '
                                   '  '
                                   'May contain:  '
                                   '* origin  '
                                   '* country  '
                                   '* city  '
                                   '* place  '
                                   '* user_groups  '
                                   '* user  '
                                   '* post  '
                                   '* post_reaction  '
                                   '  '
                                   'Notes:  '
                                   '* Excluded objects will not be queried, but empty '
                                   'objects may be created due to referenced '
                                   'foreign key relationships.  '
                                   '* Defaults to '
                                   '`origin,post`  '
                                   '[1]: https://lbsn.vgiscience.org/structure/',
                                   type=str)
        settings_args.add_argument("--include_lbsn_bases",
                                   help='lbsn bases to update '
                                   '  '
                                   '  '
                                   'If the target output type is `hll`, '
                                   'provide a comma separated list '
                                   'of lbsn bases to include/update/store to. '
                                   '  '
                                   '  '
                                   'Currently supported:  '
                                   '* hashtag  '
                                   '* emoji  '
                                   '* term  '
                                   '* _hashtag_latlng  '
                                   '* _term_latlng  '
                                   '* _emoji_latlng  '
                                   '* _month_hashtag  '
                                   '* _month_latlng  '
                                   '* monthofyear  '
                                   '* month  '
                                   '* dayofmonth  '
                                   '* dayofweek  '
                                   '* hourofday  '
                                   '* year  '
                                   '* month  '
                                   '* date  '
                                   '* timestamp  '
                                   '* country  '
                                   '* region  '
                                   '* city  '
                                   '* place  '
                                   '* latlng  '
                                   '* community  '
                                   '  '
                                   'Bases not included will be skipped. Per '
                                   'default, **no bases** will be considered. '
                                   '  '
                                   'Example:  '
                                   '```bash'
                                   '--include_lbsn_bases hashtag,place,date,community  '
                                   '```'
                                   'Argument only allowed one time.',
                                   type=str)
        settings_args.add_argument("--override_lbsn_query_schema",
                                   help='Override schema and table name '
                                   '  '
                                   '  '
                                   'This can be used to redirect lbsn queries on '
                                   'the given object from input db to a specific schema/table '
                                   'such as a materialized view. '
                                   '  '
                                   '  '
                                   'This can be usefull (e.g.) to limit '
                                   'processing of input data to a specific '
                                   'query.  '
                                   '  '
                                   'Format is `lbsn_type,schema.table`.  '
                                   '  '
                                   'Example:  '
                                   '```bash'
                                   '--override_lbsn_query_schema post,mviews.mypostquery  '
                                   '```'
                                   'Argument can be used multiple times.',
                                   action='append',
                                   type=str)
        settings_args.add_argument("--hmac_key",
                                   help='Override db hmac key '
                                   '  '
                                   '  '
                                   'The hmac key that is used for cryptographic hashing '
                                   'during creation of HLL sets. Override what is '
                                   'set in hllworker database here. '
                                   '  '
                                   '  '
                                   'Remember to re-use the same hmac key for any '
                                   'consecutive update of HLL sets. '
                                   '  '
                                   '  '
                                   'The crypt.salt variable can also be set (temporarily or permanently) in the '
                                   'hll worker database itself. '
                                   '  '
                                   'Example:  '
                                   '```sql'
                                   'ALTER DATABASE hllworkerdb SET crypt.salt = \'CRYPTSALT\';  '
                                   '```'
                                   '  '
                                   '  '
                                   'Further information is available in the [YFCC HLL tutorial][2]. '
                                   '  '
                                   '[2]: https://lbsn.vgiscience.org/tutorial/yfcc-geohash/#prepare-query-and-cryptographic-hashing',
                                   action='append',
                                   type=str)

        args = parser.parse_args()
        if args.file_input:
            self.is_local_input = True
            self.local_file_type = args.file_type
            if args.is_stacked_json:
                self.is_stacked_json = True
            if args.is_line_separated_json:
                self.is_line_separated_json = True
            if not args.input_path_url:
                self.input_path = Path.cwd() / "01_Input"
                print(f'Using Path: {self.input_path}')
            else:
                if str(args.input_path_url).startswith('http'):
                    self.input_path = args.input_path_url.split(";")
                    self.source_web = True
                else:
                    input_path = Path(args.input_path_url)
                    self.input_path = input_path
        else:
            self.dbuser_input = args.dbuser_input
            self.dbpassword_input = args.dbpassword_input
            input_ip, input_port = BaseConfig.get_ip_port(
                args.dbserveraddress_input)
            self.dbserveraddress_input = input_ip
            if input_port:
                self.dbserverport_input = input_port
            self.dbname_input = args.dbname_input
        if args.dry_run:
            self.dry_run = True
        if args.csv_output:
            raise NotImplementedError(
                "CSV output is currently not available.")
        if args.dbformat_input:
            self.dbformat_input = args.dbformat_input
        if args.origin:
            self.origin = args.origin
        if args.geocode_locations:
            self.geocode_locations = Path(
                args.geocode_locations)
        if args.ignore_input_source_list:
            self.ignore_input_source_list = Path(
                args.ignore_input_source_list)
        if not self.csv_output and args.dbuser_output:
            self.dbuser_output = args.dbuser_output
            self.dbpassword_output = args.dbpassword_output
            output_ip, output_port = BaseConfig.get_ip_port(
                args.dbserveraddress_output)
            self.dbserveraddress_output = output_ip
            if output_port:
                self.dbserverport_output = output_port
            self.dbname_output = args.dbname_output
        if not self.csv_output and args.dbformat_output:
            self.dbformat_output = args.dbformat_output
            if self.dbformat_output == "hll":
                try:
                    self.dbname_hllworker = args.dbname_hllworker
                    self.dbuser_hllworker = args.dbuser_hllworker
                    self.dbpassword_hllworker = args.dbpassword_hllworker
                    hllworker_ip, hllworker_port = BaseConfig.get_ip_port(
                        args.dbserveraddress_hllworker)
                    self.dbserveraddress_hllworker = hllworker_ip
                    if hllworker_port:
                        self.dbserverport_hllworker = hllworker_port
                except ValueError:
                    raise ValueError(
                        "If '--dbformat_output hll' is used, also provide "
                        "a HLL Worker Connection (--dbname_hllworker, "
                        "--dbuser_hllworker, --dbpassword_hllworker, "
                        "--dbserveraddress_hllworker)")
        if args.transferlimit:
            self.transferlimit = args.transferlimit
            if self.transferlimit == 0:
                self.transferlimit = None
        if args.transfer_count:
            self.transfer_count = args.transfer_count
        if args.records_tofetch:
            self.number_of_records_to_fetch = args.records_tofetch
        if args.disable_transfer_reactions:
            self.transfer_reactions = False
        if args.disable_reaction_post_referencing:
            self.disable_reactionpost_ref = True
        if args.ignore_non_geotagged:
            self.ignore_non_geotagged = True
        if args.startwith_db_rownumber:
            # hack: converts to int (row-count) if possible
            # otherwise uses str (row-id lookup)
            # TODO: separate into different cli args
            self.startwith_db_rownumber = self.int_or_str(
                args.startwith_db_rownumber)
        if args.endwith_db_rownumber:
            self.endwith_db_rownumber = args.endwith_db_rownumber
        if args.debug_mode:
            self.debug_mode = args.debug_mode
        if args.input_lbsn_type:
            self.input_lbsn_type = args.input_lbsn_type
        if args.map_full_relations:
            self.map_relations = True
        if args.csv_allow_linebreaks:
            self.csv_suppress_linebreaks = False
        if args.csv_delimiter:
            self.csv_delim = args.csv_delimiter
        if args.use_csv_dictreader:
            self.use_csv_dictreader = args.use_csv_dictreader
        if args.recursive_load:
            self.recursive_load = True
        if args.skip_until_file:
            self.skip_until_file = args.skip_until_file
        if args.zip_records:
            self.zip_records = True
        if args.skip_until_record:
            self.skip_until_record = args.skip_until_record
        if args.mappings_path:
            self.mappings_path = Path(args.mappings_path)
        if args.min_geoaccuracy:
            self.min_geoaccuracy = self.check_geoaccuracy_input(
                args.min_geoaccuracy)
        if args.include_lbsn_objects:
            self.include_lbsn_objects = \
                args.include_lbsn_objects.lower().split(",")
        else:
            self.include_lbsn_objects = ['origin,post']
        if args.include_lbsn_bases:
            self.include_lbsn_bases = args.include_lbsn_bases.split(",")
        if args.override_lbsn_query_schema:
            self.override_lbsn_query_schema = self.compile_schema_override(
                args.override_lbsn_query_schema)
        if args.hmac_key:
            self.hmac_key = args.hmac_key

    @classmethod
    def compile_schema_override(
            cls, override_lbsn_query_schema: List[str]) -> List[Tuple[str, str]]:
        """Split lbsn_type from schema.table override"""
        lbsn_query_schema = []
        for override in override_lbsn_query_schema:
            try:
                lbsn_type, schema_table_override = override.split(
                    ",")
            except ValueError as e:
                raise ValueError(
                    f"Cannot split lbsn_type from schema.table ({override}). "
                    f"Make sure override_lbsn_query_schema entries are "
                    f"formatted correctly, e.g. lbsn_type,schema.table") from e
            lbsn_query_schema.append(
                (lbsn_type.lower(), schema_table_override.lower()))
        return lbsn_query_schema

    @staticmethod
    def check_geoaccuracy_input(geoaccuracy_string):
        """Checks geoaccuracy input string and matches
        against proto buf spec
        """
        if geoaccuracy_string == 'latlng':
            return lbsn.Post.LATLNG
        if geoaccuracy_string == 'place':
            return lbsn.Post.PLACE
        if geoaccuracy_string == 'city':
            return lbsn.Post.CITY
        print("Unknown geoaccuracy.")
        return None

    @staticmethod
    def get_ip_port(ip_string: str):
        """Gets IP and port, if available

        Arguments:
            ip_string {str} -- String containing IPV4 and
            possibly port attached with : character
        """
        port = None
        ip_addr = None
        ip_string_spl = ip_string.split(":")
        if len(ip_string_spl) == 2:
            ip_addr = ip_string_spl[0]
            port = ip_string_spl[1]
        else:
            ip_addr = ip_string
        return ip_addr, port

    @staticmethod
    def set_options():
        """Includes global options in other packages to be set
        prior execution"""
        # tell shapely to include the srid when generating WKBs
        geos.WKBWriter.defaults['include_srid'] = True

    @classmethod
    def int_or_str(cls, value_str: str):
        """Returns int if value is of type int, otherwise str"""
        try:
            return int(value_str)
        except:
            return value_str



# -*- coding: utf-8 -*-

"""
LBSNTransform: Convert Raw Social Media data
         to common LBSN interchange format (ProtoBuf) and
         transfer to local CSV or LBSN Postgres Database
"""

from __future__ import absolute_import

__author__ = "Alexander Dunkel"
__license__ = "GNU GPLv3"

import io
import logging
import sys
from pathlib import Path

from lbsntransform.tools.helper_functions import HelperFunctions as HF
from lbsntransform.output.shared_structure import LBSNRecordDicts
from lbsntransform.output.submit_data import LBSNTransfer
from lbsntransform.input.load_data import LoadData


class LBSNTransform():
    """Import, convert and export RAW Location Based Social Media data,
    such as Twitter and Flickr, based on a common data structure concept
    in Google's ProtoBuf format (see package lbsnstructure).

    Input can be:
        - local CSV or Json (stacked/regular/line separated)
        - Postgres DB connection
    Output can be:
        - local CSV
        - local file with ProtoBuf encoded records
        - local SQL file ready for "Import from" in Postgres LBSN db
        - Postgres DB connection (with existing LBSN DB Structure)

    Parameters
    ----------

    origin_id : int, optional (default=3)
        Type of input source. Each input source has its own import mapper
        defined in a class. Feel free to add or modify classes based
        on your needs. Pre-provided are:
            2    - Flickr
            2.1  - Flickr YFCC100M dataset
            3    - Twitter
    """

    def __init__(
            self, importer, logging_level=None,
            is_local_input: bool = False, transfer_count: int = 50000,
            csv_output: bool = True, csv_suppress_linebreaks: bool = True,
            dbuser_output=None, dbserveraddress_output=None, dbname_output=None,
            dbpassword_output=None, dbserverport_output=None,
            dbuser_input=None, dbserveraddress_input=None, dbname_input=None,
            dbpassword_input=None, dbserverport_input=None,
            dbformat_output=None, dbuser_hllworker=None,
            dbserveraddress_hllworker=None, dbname_hllworker=None,
            dbpassword_hllworker=None, dbserverport_hllworker=None,
            include_lbsn_bases=None, dry_run=None, hmac_key=None):
        """Init settings for LBSNTransform"""

        # init logger level
        if logging_level is None:
            logging_level = logging.INFO
        # Set Output to Replace in case of encoding issues (console/windows)
        sys.stdout = io.TextIOWrapper(
            sys.stdout.detach(), sys.stdout.encoding, 'replace')
        sys.stdout.flush()
        self.log = HF.set_logger()
        self.dry_run = dry_run
        # init global settings

        self.transfer_count = transfer_count
        self.importer = importer
        # get origin name and id from importer
        # e.g. yfcc100m dataset has origin id 21,
        # but is specified as general Flickr origin (2) in importer
        self.origin_id = self.importer.ORIGIN_ID
        self.origin_name = self.importer.ORIGIN_NAME
        # establish output connection
        self.dbuser_output = dbuser_output
        conn_output, cursor_output = LoadData.initialize_connection(
            dbuser_output, dbserveraddress_output,
            dbname_output, dbpassword_output, dbserverport_output)
        if dbformat_output == "hll":
            __, cursor_hllworker = LoadData.initialize_connection(
                dbuser_hllworker, dbserveraddress_hllworker,
                dbname_hllworker, dbpassword_hllworker, dbserverport_hllworker,
                readonly=True)
        else:
            cursor_hllworker = None

        # store global for closing connection later
        self.cursor_output = cursor_output
        self.output = LBSNTransfer(
            db_cursor=cursor_output,
            db_connection=conn_output,
            store_csv=csv_output,
            SUPPRESS_LINEBREAKS=csv_suppress_linebreaks,
            dbformat_output=dbformat_output,
            hllworker_cursor=cursor_hllworker,
            include_lbsn_bases=include_lbsn_bases,
            dry_run=self.dry_run)
        # load from local json/csv or from PostgresDB
        self.cursor_input = None
        self.is_local_input = is_local_input
        if not self.is_local_input:
            __, cursor_input = LoadData.initialize_connection(
                dbuser_input, dbserveraddress_input,
                dbname_input, dbpassword_input, dbserverport_input,
                readonly=True, dict_cursor=True)
            self.cursor_input = cursor_input

        # initialize stats
        self.processed_total = 0
        self.initial_loop = True
        self.how_long = None
        # field mapping structure
        # this is where all the converted data will be stored
        # note that one input record may contain many lbsn records
        self.lbsn_records = LBSNRecordDicts()
        # prepare hmac
        if self.output.dbformat_output == 'hll':
            self.output.prepare_hmac(
                hmac_key)

    def add_processed_records(self, lbsn_record):
        """Adds one or multiple LBSN Records (ProtoBuf)
        to collection (dicts of LBSNRecords)

        Will automatically call self.store_lbsn_records()
        """
        self.lbsn_records.add_records_to_dict(
            lbsn_record)
        self.processed_total += 1
        # On the first loop
        # or after 50.000 (default) processed records,
        # store results
        if self.initial_loop:
            if self.output.dbformat_output == 'lbsn':
                self.output.store_origin(
                    self.origin_id, self.origin_name)
                self.store_lbsn_records()
            self.initial_loop = False
        if self.lbsn_records.count_glob >= self.transfer_count:
            print("\n", end='')
            self.store_lbsn_records()

    def store_lbsn_records(self):
        """Stores processed LBSN Records to chosen output format
        """
        self.output.store_lbsn_record_dicts(
            self.lbsn_records)
        self.output.commit_changes()
        self.lbsn_records.clear()

    def finalize_output(self):
        """finalize all transactions (csv merge etc.)
        """
        self.store_lbsn_records()
        self.output.finalize()
        # Close connections to DBs
        if not self.is_local_input:
            self.cursor_input.close()
        if self.dbuser_output:
            self.cursor_output.close()

    @staticmethod
    def close_log():
        """"Closes log and writes to archive file
        """
        logging.shutdown()
        # rename log file for archive purposes
        today = HF.get_str_formatted_today()
        outfile = Path(f"{today}.log")
        with open(outfile, 'a+') as outfile:
            with open('log.log') as infile:
                outfile.write(f'\n')
                for line in infile:
                    outfile.write(line)



# -*- coding: utf-8 -*-
"""Provides functions to connect to Postgres DB

"""

import datetime
import getpass
import logging
import sys

import psycopg2
import psycopg2.extras
from psycopg2.extras import DictCursor

LOG = logging.getLogger()


class DBConnection():
    """ Class for connectiong to Postgres. """

    def __init__(self, serveraddress=None, dbname=None,
                 user=None, password=0, readonly=False,
                 sslmode='prefer', port=5432):
        """Initialize DBConnection object with attributes, if passed. """
        self.serveraddress = serveraddress
        self.dbname = dbname
        self.user = user
        self.password = password
        self.readonly = readonly
        self.sslmode = sslmode  # Choose 'disable' to connect without ssl
        self.port = port

    def connect(self, dict_cursor=None):
        """Database config. """
        conf = {
            "host": self.serveraddress,
            "dbname": self.dbname,
            "user": self.user,
            "sslmode": self.sslmode,
            "port": self.port
        }
        # Connect to database
        if self.password == 0:
            promt_txt = f"Enter the password for {conf['user']}: "
            if sys.stdin.isatty():
                conf["password"] = getpass.getpass(prompt=promt_txt)
            else:
                print(promt_txt)
                conf["password"] = sys.stdin.readline().rstrip()
        else:
            conf["password"] = self.password
        # define connection string
        conn_string = f"host='{conf['host']}'" \
                      f"dbname='{conf['dbname']}'" \
                      f"user='{conf['user']}'" \
                      f"password='{conf['password']}'" \
                      f"sslmode='{conf['sslmode']}'" \
                      f"port='{conf['port']}'" \
                      f"application_name='LBSN Batch Transfer'"
        cursor_factory = None
        if dict_cursor:
            cursor_factory = DictCursor
        # get a connection, if a connect cannot be made an
        # exception will be raised here
        try:
            conn = psycopg2.connect(conn_string, cursor_factory=cursor_factory)
        except Exception as err:
            print(err)
            sys.exit()
        # activate dict to hstore conversion globally
        psycopg2.extras.register_hstore(conn, globally=True)
        # conn.cursor will return a cursor object,
        # this will be used to perform queries
        cursor = conn.cursor()
        dnow = datetime.datetime.now()
        LOG.info(f'{dnow.strftime("%Y-%m-%d %H:%M:%S")} '
                 f'- Connected to {self.dbname}')
        return conn, cursor



"""Script that does final markdown formatting 
   after argdown parse_args to markdown conversion
   
   Note:
   It is recognized that this is a horrible way to do things.
   We'll use this until an arg parser extension is available 
   for mkdocs.
"""

import re

def format_markdown():
    """Final markdown formatting fixes
    """
    # open file to output source code
    source_file_path = "docs/argparse/args.md"
    # fix markdown linebreaks from python docstrings convert
    newlines = []
    with open(source_file_path, "r") as source_file:
        for line in source_file:
            newline = line.replace('&nbsp;&nbsp;<br>', '  \n')
            newline = newline.replace('!!!\n', '!!! ')
            newline = newline.replace('&nbsp;&nbsp;&nbsp;&nbsp;', '\n    ')
            newline = newline.replace('```', '\n```\n')
            newline = newline.replace(':  \n', ':  \n\n')
            newline = newline.replace(']:\n', ']: ')
            newline = newline.replace('\n*\n','\n* ')
            newlines.append(newline)
    all_lines = ''.join(newlines)
    # fix code blocks from argdown
    # first remove all line breaks
    # and add valid ones afterwards
    pat = r'```(.*?)```'
    regex = re.compile(pat, re.DOTALL|re.MULTILINE)
    # get number of total matches
    match_count = len([x for x in re.finditer(regex, all_lines)])
    for match_x in range(match_count-1):
        for x, match in enumerate(re.finditer(regex, all_lines)):
            if x == 0:
                # skip first code blocks
                # with index
                continue
            if x != match_x+1:
                continue
            # Start index of match (integer)
            sStart = match.start()
        
            # Final index of match (integer)
            sEnd = match.end()
            
            # Complete match (string)
            sGroup = match.group().replace(
                '\n'," ").replace(
                    '  ',"\n").replace(
                                '``` ','```').replace(
                                    '-- ','--').replace(
                                        '```bash',"```bash\n").replace(
                                            '```python',"```python\n").replace(
                                                '```sql',"```sql\n").replace(
                                                    '\n ',"\n").replace(
                                                        '\n\n',"\n")
            all_lines="".join((all_lines[:sStart],sGroup,all_lines[sEnd:]))
            # Print match
            # print('Match "{}" found at: [{},{}]'.format(sGroup, sStart,sEnd))  
    # cleanup
    with open("docs/argparse/args.md", "w") as source_file:
        source_file.write(all_lines)

# run script
format_markdown()



# -*- coding: utf-8 -*-

"""Hll bases for social facet
"""

from typing import Tuple, Union

from lbsnstructure import lbsnstructure_pb2 as lbsn
from lbsntransform.output.hll import hll_bases as hll

from lbsntransform.output.hll.hll_functions import HLLFunctions as HLF

FACET = 'social'


class SocialBase(hll.HllBase):
    """Intermediate social base class that extends HllBase
    for common social base attributes
    """

    def __init__(self):
        super().__init__()
        self.metrics['latlng_hll'] = set()
        self.metrics['place_hll'] = set()
        self.metrics['date_hll'] = set()


class UserBase(SocialBase):
    """Extends Social Base Class"""
    NAME = hll.HllBaseRef(facet=FACET, base='user')

    def __init__(self, record: lbsn.User = None):
        super().__init__()
        self.key['user_guid'] = None
        if record is None:
            # init empty
            return
        # TODO: implement one-way-hashing
        self.key['user_guid'] = HLF.hll_concat_origin_guid(record)


class FriendsBase(SocialBase):
    """Extends Social Base Class"""
    NAME = hll.HllBaseRef(facet=FACET, base='friends')

    def __init__(self, friends_record: Tuple[lbsn.User, lbsn.User] = None):
        super().__init__()
        self.key['user_guid'] = None
        self.key['user_guid_friend'] = None
        if friends_record is None:
            # init empty
            return
        # TODO: implement one-way-hashing
        self.key['user_guid'] = HLF.hll_concat_origin_guid(
            friends_record[0])
        self.key['user_guid_friend'] = HLF.hll_concat_origin_guid(
            friends_record[1])


class CommunityBase(SocialBase):
    """Extends Social Base Class"""
    NAME = hll.HllBaseRef(facet=FACET, base='community')

    def __init__(self, record: Union[lbsn.Origin, lbsn.Post] = None):
        super().__init__()
        self.attrs['name'] = None
        self.key['id'] = None
        if record is None:
            # init empty
            return
        if hasattr(record, "origin_id"):
            self.key['id'] = record.origin_id
            self.attrs['name'] = record.DESCRIPTOR.name
        else:
            self.key['id'] = record.pkey.origin.origin_id


class CultureBase(SocialBase):
    """Extends Social Base Class"""
    NAME = hll.HllBaseRef(facet=FACET, base='culture')

    def __init__(self, record: lbsn.Language = None):
        super().__init__()
        self.key['id'] = None
        if record is None:
            # init empty
            return
        self.key['id'] = record.language_short



# -*- coding: utf-8 -*-

"""
Module for db input connection sql mapping
"""

import enum
from typing import Union, Optional, List, Tuple
from lbsnstructure import lbsnstructure_pb2 as lbsn

"""Schema convention from lbsn db spec"""
LBSN_SCHEMA = [
    (lbsn.Origin().DESCRIPTOR.name, "social", "origin", "origin_id"),
    (lbsn.Country().DESCRIPTOR.name, "spatial", "country", "country_guid"),
    (lbsn.City().DESCRIPTOR.name, "spatial", "city", "city_guid"),
    (lbsn.Place().DESCRIPTOR.name, "spatial", "place", "place_guid"),
    (lbsn.UserGroup().DESCRIPTOR.name, "social", "user_groups", "usergroup_guid"),
    (lbsn.User().DESCRIPTOR.name, "social", "user", "user_guid"),
    (lbsn.Post().DESCRIPTOR.name, "topical", "post", "post_guid"),
    (lbsn.PostReaction().DESCRIPTOR.name,
     "topical", "post_reaction", "reaction_guid"),
]


def optional_schema_override(
        LBSN_SCHEMA: List[Tuple[str, str, str, str]],
        schema_table_overrides: List[Tuple[str, str]]) -> List[Tuple[str, str, str, str]]:
    """Override schema and table name for selected lbsn objects."""
    LBSN_SCHEMA_OVERRIDE = []
    for lbsn_type, schema_name, table_name, key_col in LBSN_SCHEMA:
        for schema_table_override in schema_table_overrides:
            lbsn_object_ref, schema_table_override = schema_table_override
            try:
                schema_override, table_override = schema_table_override.split(
                    ".")
            except ValueError as e:
                raise ValueError(
                    f"Cannot split schema and table from override "
                    f"({schema_table_override}). Make sure "
                    f"override_lbsn_query_schema entries are formatted "
                    f"correctly, e.g. schema_override.table_override") from e
            if lbsn_type.lower() == lbsn_object_ref:
                # append override
                LBSN_SCHEMA_OVERRIDE.append(
                    (lbsn_type, schema_override, table_override, key_col))
                break
        else:
            # append none-overrides
            # if no match has been found
            LBSN_SCHEMA_OVERRIDE.append(
                (lbsn_type, schema_name, table_name, key_col))
    return LBSN_SCHEMA_OVERRIDE


class InputSQL(enum.Enum):
    """SQL for default JSON records stored in DB

    In this example, records are stored in table "input" in schema
    public. They're 3 columns, in_id, insert_time and data.
    Table data is cast to json type.

    The two %s string formatters allow substitution of values
    during the query. The first %s is the key, the second %s is the limit of
    records to get during the query (default 10000)
    """
    DEFAULT = '''
            SELECT in_id, insert_time, data::json
            FROM {schema_name}."{table_name}"
            WHERE {key_col} > {start_id}
            ORDER BY {key_col} ASC
            LIMIT {number_of_records_to_fetch};
            '''

    """SQL for LBSN records stored in DB"""
    LBSN = '''
            SELECT * FROM {schema_name}."{table_name}"
            {optional_where}
            ORDER BY {key_col} ASC
            LIMIT {number_of_records_to_fetch};
            '''

    DB_CUSTOM = '''Define your own DB Mapping SQL here'''

    def get_sql(
            self, schema_name: str = "public", table_name: str = "input",
            start_id: Optional[Union[int, str]] = None,
            number_of_records_to_fetch: int = 10000,
            key_col="in_id"):
        """Get SQL formatted string"""
        optional_where = ""
        if start_id is not None:
            quote_subst = ""
            if self.name == "LBSN":
                # quoted string required
                quote_subst = "'"
            optional_where = \
                f"WHERE {key_col} > {quote_subst}{start_id}{quote_subst}"
        # self.value refers to current ENUM,
        # which is always string
        return self.value.format(
            schema_name=schema_name,
            table_name=table_name,
            optional_where=optional_where,
            number_of_records_to_fetch=number_of_records_to_fetch,
            key_col=key_col)



# -*- coding: utf-8 -*-

"""
Module for mapping Flickr YFCC100M dataset to common LBSN Structure.
"""

import csv
import logging
import re
from decimal import Decimal
from urllib.parse import unquote
from typing import List, Optional

from lbsnstructure import lbsnstructure_pb2 as lbsn

from lbsntransform.tools.helper_functions import HelperFunctions as HF

# pylint: disable=no-member

MAPPING_ID = 21

# module level constants
# for FLickr yfcc100m mapping
FLICKR_COUNTRY_MATCH = {
    "country", "timezone", "state", "territory",
    "state/territory", "land", "arrondissement", "prefecture",
    "island", "disputed territory", "overseas collectivity",
    "overseas region", "island group", "kingdom"}
FLICKR_CITY_MATCH = {
    "city", "town", "region", "special administrative region",
    "county", "district", "zip", "province",
    "district/county", "autonomous community", "municipality",
    "department", "district/town/township", "township",
    "historical town", "governorate", "commune", "canton",
    "muhafazah", "local government area", "division", "periphery",
    "zone", "sub-region", "district/town/county", "sub-district",
    "parish", "federal district", "neighborhood", "chome", "ward",
    "sub-division", "community", "autonomous region",
    "municipal region", "emirate", "autonomous province",
    "historical county", "constituency", "entity", "colloquial",
    "circle", "quarter", "dependency", "sub-prefecture", "insular area",
    "island council"}
FLICKR_PLACE_MATCH = {
    "poi", "land feature", "estate", "airport", "drainage",
    "miscellaneous", "suburb"}


class importer():
    """ Provides mapping function from Flickr YFCC100m to
        protobuf lbsnstructure
    """
    ORIGIN_NAME = "Flickr yfcc100m"
    ORIGIN_ID = 2

    def __init__(self,
                 disable_reaction_post_referencing=False,
                 geocodes=False,
                 map_full_relations=False,
                 map_reactions=True,
                 ignore_non_geotagged=False,
                 ignore_sources_set=None,
                 min_geoaccuracy=None):
        # We're dealing with Flickr in this class, lets create the OriginID
        # globally
        # this OriginID is required for all CompositeKeys
        origin = lbsn.Origin()
        origin.origin_id = lbsn.Origin.FLICKR
        self.origin = origin
        self.null_island = 0
        self.log = logging.getLogger('__main__')  # get the main logger object
        self.skipped_count = 0
        self.skipped_low_geoaccuracy = 0
        # some records in YFCC100m are larger
        # than the default csv limit in python
        # of 131072
        csv.field_size_limit(500000)
        # self.disableReactionPostReferencing = disableReactionPostReferencing
        # self.mapFullRelations = mapFullRelations
        # self.geocodes = geocodes
        self.lic_dict = {
            "All Rights Reserved": 0,
            "Attribution-NonCommercial-ShareAlike License": 1,
            "Attribution-NonCommercial License": 2,
            "Attribution-NonCommercial-NoDerivs License": 3,
            "Attribution License": 4,
            "Attribution-ShareAlike License": 5,
            "Attribution-NoDerivs License": 6,
            "No known copyright restrictions": 7,
            "United States Government Work": 8,
            "Public Domain Dedication (CC0)": 9,
            "Public Domain Mark": 10
        }

    def parse_csv_record(self, record, record_type: Optional[str] = None):
        """Entry point for flickr CSV data:
            - Decide if CSV record contains place-info or post-info
            - Skip empty or malformed records

        Attributes:
        record    A single row from CSV, stored as list type.
        """
        # Flickr photo CSV has length of 25 columns
        # if more, then we're dealing with concat photo+place pipe
        if len(record) == 2:
            # only YFCC100m place record
            lbsn_records_place = self.extract_flickr_place(record)
            return lbsn_records_place
        if len(record) < 12 or not record[0].isdigit():
            # skip erroneous entries + log
            self.skipped_count += 1
            return
        # process regular Flickr yfcc100m record
        lbsn_records = self.extract_flickr_post(record)
        return lbsn_records

    def extract_flickr_post(self, record):
        """Main function for processing Flickr YFCC100M CSV entry.
           This mothod is adapted to a special structure, adapt if needed.

        To Do:
            - parameterize column numbers and structure
            - provide external config-file for specific CSV structures
            - currently not included in lbsn mapping are MachineTags,
              GeoContext (indoors, outdoors), WoeId
              and some extra attributes only present for Flickr

        Overview of available columns and examples:
        0 row-number    -   0
        1 Photo/video identifier    -   6985418911
        2 lbsn.User NSID(PostID?)     -   4e2f7a26a1dfbf165a7e30bdabf7e72a
        3 lbsn.User ID     -   39089491@N00
        4 lbsn.User nickname     -   gnuckx
        5 Date taken    -   2012-02-16 09:56:37.0
        6 Date uploaded     -   1331840483
        7 Capture device    -   Canon+PowerShot+ELPH+310+HS
        8 Title     -   IMG_0520
        9 Description      -     My vacation
        10 tags (comma-separated)   -   canon,canon+powershot+hs+310
        11 Machine tags (comma-separated)   - landscape, hills, water
        12 Longitude    -   -81.804885
        13 Latitude     -   24.550558
        14 Accuracy -   12
        15 Photo/video page URL -   http://www.flickr.com/photos/39089491@N00/6985418911/
        16 Photo/video download URL -   http://farm8.staticflickr.com/7205/6985418911_df7747990d.jpg
        17 License name -   Attribution-NonCommercial-NoDerivs License
        18 License URL  -   http://creativecommons.org/licenses/by-nc-nd/2.0/
        19 Photo/video server identifier    -   7205
        20 Photo/video farm identifier  -   8
        21 Photo/video secret   -   df7747990d
        22 Photo/video secret original  -   692d7e0a7f
        23 Extension of the original photo  -   jpg
        24 Marker (0 ¼ photo, 1 ¼ video)    -   0
        if concat:
            25 Photo/video identifier
            26 Place references (null to multiple)
        """
        # note that one input record may contain many lbsn records
        # therefore, return list of processed records
        lbsn_records = []
        # start mapping input to lbsn_records
        post_guid = record[1]
        if not HF.check_notice_empty_post_guid(post_guid):
            return None
        post_record = HF.new_lbsn_record_with_id(lbsn.Post(),
                                                 post_guid,
                                                 self.origin)
        user_record = HF.new_lbsn_record_with_id(lbsn.User(),
                                                 record[3],
                                                 self.origin)
        user_record.user_name = unquote(record[4]).replace(
            '+', ' ')
        user_record.url = f'http://www.flickr.com/photos/{user_record.pkey.id}/'
        if user_record:
            post_record.user_pkey.CopyFrom(user_record.pkey)
        lbsn_records.append(user_record)
        post_record.post_latlng = self.flickr_extract_postlatlng(record)
        geoaccuracy = importer.flickr_map_geoaccuracy(
            record[14])
        if geoaccuracy:
            post_record.post_geoaccuracy = geoaccuracy
        # place record available in separate yfcc100m dataset
        # if record[1]:
            # we need some information from postRecord to create placeRecord
            # (e.g.  user language, geoaccuracy, post_latlng)
            # some of the information from place will also modify postRecord
            # place_record = HF.new_lbsn_record_with_id(lbsn.Place(),
            #                                           record[1],
            #                                           self.origin)
            # lbsn_records.append(place_record)
            # post_record.place_pkey.CopyFrom(place_record.pkey)
        post_record.post_publish_date.CopyFrom(
            HF.parse_timestamp_string_to_protobuf(record[6]))
        post_created_date = HF.parse_csv_datestring_to_protobuf(
            record[5], t_format='%Y-%m-%d %H:%M:%S.%f')
        if post_created_date:
            post_record.post_create_date.CopyFrom(
                post_created_date
            )
        post_record.post_views_count = 0
        post_record.post_comment_count = 0
        post_record.post_like_count = 0
        post_record.post_url = record[15]
        # YFCC100M dataset contains HTML codes (%20) and
        # space character is replaced by +
        post_record.post_body = unquote(record[9]).replace(
            '+', ' ')
        post_record.post_title = unquote(record[8]).replace(
            '+', ' ')
        post_record.post_thumbnail_url = record[16]  # note: fullsize url!
        # split tags by , and + because by lbsn-spec,
        # tags are limited to single word
        record_tags_list = list(
            set(filter(
                None,
                [HF.remove_prefix(unquote(tag), "#")
                 for tag in re.split("[,+]+", record[10])]
            )))
        if record_tags_list:
            for tag in record_tags_list:
                tag = importer.clean_tags_from_flickr(tag)
                post_record.hashtags.append(tag)
        record_machine_tags = list(
            set(filter(None, [unquote(mtag) for mtag in re.split("[,+]+", record[11])])))
        if 'video' in record_machine_tags:
            # all videos appear to have 'video' in machine tags
            post_record.post_type = lbsn.Post.VIDEO
        else:
            post_record.post_type = lbsn.Post.IMAGE
        # replace text-string of content license by integer-id
        if record[17] is not None:
            post_record.post_content_license = \
                self.get_license_number_from_license_name(record[17])
        # place record available in separate yfcc100m dataset
        # if records parsed as joined urls, length is larger than 25
        if len(record) > 25:
            post_plus_place_records = self.extract_flickr_place(
                record[25:], post_record=post_record)
            if post_plus_place_records is None:
                lbsn_records.append(post_record)
            else:
                lbsn_records.extend(post_plus_place_records)
        else:
            lbsn_records.append(post_record)
        return lbsn_records

    def extract_flickr_place(self, record, post_record: lbsn.Post = None):
        """Main function for processing Flickr YFCC100M lbsn.Place CSV entry
        lbsn.Place records are available from a separate yfcc100m dataset
        and are joined based on photo guid
        """
        # note that one input record may contain many lbsn records
        # therefore, return list of processed records
        lbsn_records = []
        # start mapping input to lbsn_records
        post_guid = record[0]
        if post_record is not None:
            # verify matching guids
            if not post_record.pkey.id == post_guid:
                raise ValueError(
                    f"Post_guids not in sync: {post_record.pkey.id} "
                    f"/ {post_guid}")
        if not record[1]:
            # skip empty records
            return None
        # place records in yfcc100m contain multiple entries, e.g.
        # 24703176:Admiralty:Suburb,24703128:Central+and+Western:Territory,
        # 24865698:Hong+Kong:Special Administrative Region,
        # 28350827:Asia%2FHong_Kong:Timezone
        place_records = record[1].split(",")
        for place_record in place_records:
            lbsn_place_record = \
                importer.process_place_record(
                    place_record, self.origin)
            lbsn_records.append(lbsn_place_record)
        # create or update post record with entries from place record
        post_record = self.update_post_with_place(
            post_record=post_record,
            post_guid=post_guid,
            place_records=lbsn_records)
        lbsn_records.append(post_record)
        return lbsn_records

    def update_post_with_place(
            self, post_record: lbsn.Post = None, post_guid: str = None,
            place_records: List[lbsn.Place] = None):
        """Update post record with entries from place record"""
        if post_record is None:
            if post_guid is None:
                raise ValueError("Cannot create lbsn.Post without post_guid")
            # create new post record
            post_record = HF.new_lbsn_record_with_id(
                lbsn.Post(), post_guid, self.origin)
        if place_records is None:
            return post_record
        for place_record in place_records:
            if isinstance(place_record, lbsn.Country):
                post_record.country_pkey.CopyFrom(place_record.pkey)
            if isinstance(place_record, lbsn.City):
                post_record.city_pkey.CopyFrom(place_record.pkey)
            # either city or place, Twitter user cannot attach both (?)
            elif isinstance(place_record, lbsn.Place):
                post_record.place_pkey.CopyFrom(place_record.pkey)
        return post_record

    @staticmethod
    def process_place_record(place_record, origin):
        """Assignment of Flickr place types to lbsnstructure
        hierarchy: lbsn.Country, lbsn.City, lbsn.Place
        Original Flickr place types, which are more detailed,
        are stored in sub_type field
        """
        place_record_split = place_record.split(":")
        if not len(place_record_split) == 3:
            raise ValueError(
                f'Malformed place entry:\n'
                f'place_record: {place_record}')
        place_guid = unquote(place_record_split[0])
        place_name = unquote(place_record_split[1]).replace(
            '+', ' ')
        place_type = unquote(place_record_split[2])
        place_type_lw = place_type.lower()
        place_type_lw_split = place_type_lw.split("/")
        # assignment
        if any(ptls in FLICKR_COUNTRY_MATCH for ptls in place_type_lw_split):
            lbsn_place_record = HF.new_lbsn_record_with_id(
                lbsn.Country(), place_guid, origin)
        elif any(ptls in FLICKR_CITY_MATCH for ptls in place_type_lw_split):
            lbsn_place_record = HF.new_lbsn_record_with_id(
                lbsn.City(), place_guid, origin)
        elif any(ptls in FLICKR_PLACE_MATCH for ptls in place_type_lw_split):
            lbsn_place_record = HF.new_lbsn_record_with_id(
                lbsn.Place(), place_guid, origin)
        else:
            logging.getLogger('__main__').debug(
                f'Could not assign place type {place_type_lw}\n'
                f'found in place_record: {place_record}\n'
                f'Will assign default "lbsn.Place"')
            lbsn_place_record = HF.new_lbsn_record_with_id(
                lbsn.Place(), place_guid, origin)
        lbsn_place_record.name = place_name
        if isinstance(lbsn_place_record, lbsn.City):
            # record sub types only for city and place
            lbsn_place_record.sub_type = place_type
        elif isinstance(lbsn_place_record, lbsn.Place):
            lbsn_place_record.place_description = place_type
        # place_record.url (not provided)
        # need to consult post data for lat/lng coordinates
        # set to null island first
        lbsn_place_record.geom_center = "POINT(%s %s)" % (0, 0)
        return lbsn_place_record

    @staticmethod
    def clean_tags_from_flickr(tag):
        """Clean special vars not allowed in tags.
        """
        characters_to_replace = ('{', '}')
        for char_check in characters_to_replace:
            tag = tag.replace(char_check, '')
        return tag

    def get_license_number_from_license_name(self, license_name: str):
        """Flickr YFCC100M contains only full string names of licenses
        This function converts names to ids.

        Arguments:
            license_name {str} -- YFCC100M license name
        see:
        https://www.flickr.com/services/api/flickr.photos.licenses.getInfo.html
        """
        lic_number = self.lic_dict.get(license_name)
        return lic_number

    @staticmethod
    def flickr_map_geoaccuracy(flickr_geo_accuracy_level):
        """Flickr Geoaccuracy Levels (16) are mapped to four
           LBSNstructure levels (LBSN PostGeoaccuracy):
           UNKNOWN = 0; LATLNG = 1; PLACE = 2; CITY = 3; COUNTRY = 4
           Fickr: World level is 1, lbsn.Country is ~3, Region ~6,
           lbsn.City ~11, Street ~16.

           Flickr current range is 1-16. Defaults to 16 if not specified.

        Attributes:
        flickr_geo_accuracy_level   Geoaccuracy Level returned from Flickr
        (String, e.g.: "Level12")
        """
        lbsn_geoaccuracy = False
        stripped_level = flickr_geo_accuracy_level.lstrip('Level').strip()
        if stripped_level.isdigit():
            stripped_level = int(stripped_level)
            if stripped_level >= 15:
                lbsn_geoaccuracy = lbsn.Post.LATLNG
            elif stripped_level >= 12:
                lbsn_geoaccuracy = lbsn.Post.PLACE
            elif stripped_level >= 8:
                lbsn_geoaccuracy = lbsn.Post.CITY
            else:
                lbsn_geoaccuracy = lbsn.Post.COUNTRY
        else:
            if flickr_geo_accuracy_level == "Street":
                lbsn_geoaccuracy = lbsn.Post.LATLNG
            elif flickr_geo_accuracy_level in ("lbsn.City", "Region"):
                lbsn_geoaccuracy = lbsn.Post.CITY
            elif flickr_geo_accuracy_level in ("lbsn.Country", "World"):
                lbsn_geoaccuracy = lbsn.Post.COUNTRY
        return lbsn_geoaccuracy

    def flickr_extract_postlatlng(self, record):
        """Basic routine for extracting lat/lng coordinates from post.
           - checks for consistency and errors
           - in case of any issue, entry is submitted to Null island (0, 0)
        """
        lat_entry = record[13]
        lng_entry = record[12]
        if lat_entry == "" and lng_entry == "":
            l_lat, l_lng = 0, 0
        else:
            try:
                l_lng = Decimal(lng_entry)
                l_lat = Decimal(lat_entry)
            except ValueError:
                l_lat, l_lng = 0, 0
        if ((l_lat == 0 and l_lng == 0)
                or l_lat > 90 or l_lat < -90
                or l_lng > 180 or l_lng < -180):
            l_lat, l_lng = 0, 0
            self.send_to_null_island()
        return importer.lat_lng_to_wkt(l_lat, l_lng)

    @staticmethod
    def lat_lng_to_wkt(lat, lng):
        """Convert lat lng to WKT (Well-Known-Text)
        """
        point_latlng_string = "POINT(%s %s)" % (lng, lat)
        return point_latlng_string

    def send_to_null_island(self):
        """Logs entries with problematic lat/lng's,
           increases Null Island Counter by 1.
        """
        self.null_island += 1



# -*- coding: utf-8 -*-

"""
Shared structure and mapping between HLL DB and Proto LBSN Structure.
"""

# pylint: disable=no-member

from typing import Dict, List, Tuple, Union, Iterator, Optional

from lbsnstructure import lbsnstructure_pb2 as lbsn
from lbsntransform.output.hll import hll_bases as hll
from lbsntransform.output.hll.base import social, spatial, temporal, topical

from lbsntransform.output.hll.hll_functions import HLLFunctions as HLF

HllBases = Union[
    spatial.LatLngBase, spatial.PlaceBase,
    spatial.CityBase, spatial.CountryBase,
    temporal.DateBase, temporal.MonthBase,
    temporal.YearBase, temporal.MonthLatLngBase,
    temporal.MonthHashtagBase, topical.TermBase,
    topical.EmojiBase, topical.TermLatLngBase,
    topical.HashtagLatLngBase, topical.EmojiLatLngBase,
    social.CommunityBase]


class ProtoHLLMapping():
    """ Methods to map ProtoBuf structure to PG HLL SQL structure."""

    def __init__(self, include_lbsn_bases: List[str] = None):
        if include_lbsn_bases is None:
            include_lbsn_bases = []
        self.include_lbsn_bases = include_lbsn_bases

    @staticmethod
    def update_hll_dicts(
            batched_hll_bases: Dict[Tuple[str, str], Dict[str, HllBases]],
            hll_base_metrics: List[HllBases]):
        """Update batched hll bases with new list of bases, merged metrics

        batched_hll_bases:
            Dict[(facet, base)]: dict(key=str, HllBaseMetrics),
        e.g. batched_hll_bases[('spatial', 'latlng')]:
            dict[
                '71:25': user_hll, post, hll ..
                '5:16': user_hll, post, hll ..
                ]
        """
        for hll_base_metric in hll_base_metrics:
            if hll_base_metric is None:
                continue
            base_dict = batched_hll_bases.get(hll_base_metric.NAME)
            if base_dict is None:
                # skip bases not registered
                continue
            # metric_key = hll_base_metric.get_key_value()
            # metric_key is always a tuple
            # with the unique key (e.g. "(2019-09-01,)")
            # base_dict[metric_key] |= hll_base_metric
            metric_key = hll_base_metric.get_key_value()
            if metric_key not in base_dict:
                base_dict[metric_key] = hll_base_metric
            else:
                base_dict[metric_key] |= hll_base_metric

    def extract_hll_base_metrics(self, record, record_type) -> Optional[
            List[HllBases]]:
        """Combine bases and metrics from Proto LBSN record"""
        record_hll_metrics = self.get_hll_metrics(record)
        if record_hll_metrics is None:
            return
        hll_bases = self.extract_hll_bases(
            record, record_type, self.include_lbsn_bases)
        # turn generator into list
        hll_base_metrics = list(ProtoHLLMapping.update_bases_metrics(
            bases=hll_bases, metrics=record_hll_metrics))
        return hll_base_metrics

    @staticmethod
    def update_bases_metrics(
            bases: List[HllBases],
            metrics: hll.HllMetrics) -> Iterator[HllBases]:
        """Adds/updates metrics to hll_bases"""
        # iterate Namedtuple field names and values
        for key, value in zip(metrics._fields, metrics):
            # key = 'utl_hll', 'upl_hll'
            # value = set/str
            if value is None:
                continue
            for base in bases:
                if key in base.metrics:
                    if key in ('upt_hll', 'utl_hll'):
                        # use update on set
                        base.metrics[key] |= value
                    else:
                        # otherwise add single item
                        base.metrics[key].add(value)
                # return updated base as iterator
                yield base

    def extract_hll_bases(
            self, record, record_type,
            include_lbsn_bases: List[str]) -> List[HllBases]:
        """Extract hll bases from Proto LBSN record

        Depending on LBSNType, multiple bases can be extracted
        e.g.:
        lbsn.Post -> hll.LatLngBase
        lbsn.Post -> hll.PlaceBase
        lbsn.Post -> hll.CityBase
        lbsn.Post -> hll.CountryBase
        lbsn.Post -> hll.DateBase
        lbsn.Post -> hll.MonthBase
        lbsn.Post -> hll.YearBase
        lbsn.Post -> hll.TermBase (multiple)
        bsn.Post -> hll.HashtagBase (multiple)
        (lbsn.Place -> hll.PlaceBase)
        (lbsn.City -> hll.CityBase)
        """
        base_list = []
        # Posts
        if record_type == lbsn.Post.DESCRIPTOR.name:
            # Spatial Facet
            # base from record.post_geoaccuracy
            # see protobuf spec for enum values:
            # UNKNOWN = 0; LATLNG = 1; PLACE = 2; CITY = 3; COUNTRY = 4;
            spatial_base = lbsn.Post.PostGeoaccuracy.Name(
                record.post_geoaccuracy).lower()
            if include_lbsn_bases and spatial_base in include_lbsn_bases:
                base_records = self.make_base(
                    facet='spatial',
                    base=spatial_base,
                    record=record)
                if base_records:
                    base_list.extend(base_records)
            # Temporal Facet
            temporal_bases = [
                'date', 'month', 'year', '_month_latlng', '_month_hashtag']
            temporal_bases = self.filter_bases(
                temporal_bases, include_lbsn_bases)
            base_records = self.make_bases(
                facet='temporal',
                bases=temporal_bases,
                record=record)
            if base_records:
                base_list.extend(base_records)
            # Topical Facet
            topical_bases = [
                'hashtag', 'emoji', 'term', '_term_latlng', '_emoji_latlng',
                '_hashtag_latlng']
            topical_bases = self.filter_bases(
                topical_bases, include_lbsn_bases)
            base_records = self.make_bases(
                facet='topical',
                bases=topical_bases,
                record=record)
            if base_records:
                base_list.extend(base_records)
            social_bases = ['community']
            social_bases = self.filter_bases(
                social_bases, include_lbsn_bases)
            base_records = self.make_bases(
                facet='social',
                bases=social_bases,
                record=record)
            if base_records:
                base_list.extend(base_records)

        # Places
        if record_type == lbsn.Place.DESCRIPTOR.name:
            if include_lbsn_bases and 'place' in include_lbsn_bases:
                base_records = self.make_base(
                    facet='spatial',
                    base='place',
                    record=record)
                if base_records:
                    base_list.extend(base_records)
        if record_type == lbsn.City.DESCRIPTOR.name:
            if include_lbsn_bases and 'city' in include_lbsn_bases:
                base_records = self.make_base(
                    facet='spatial',
                    base='city',
                    record=record)
                if base_records:
                    base_list.extend(base_records)
        if record_type == lbsn.Country.DESCRIPTOR.name:
            if include_lbsn_bases and 'country' in include_lbsn_bases:
                base_records = self.make_base(
                    facet='spatial',
                    base='country',
                    record=record)
                if base_records:
                    base_list.extend(base_records)
        return base_list

    @classmethod
    def filter_bases(
            cls, base_list: List[str],
            include_lbsn_bases: List[str]):
        """Remove bases based on user supplied include list"""
        if not include_lbsn_bases:
            # if empty
            return base_list
        base_list_filtered = []
        for base in base_list:
            if base in include_lbsn_bases:
                base_list_filtered.append(base)
        return base_list_filtered

    @classmethod
    def make_bases(cls, facet: str, bases: List[str], record):
        """Create list of hllBases list of bases and a single lbsnRecord"""
        base_list = []
        for base in bases:
            base_records = cls.make_base(
                facet=facet,
                base=base,
                record=record)
            if base_records:
                base_list.extend(base_records)
        return base_list

    @classmethod
    def make_base(cls, facet: str, base: str, record):
        """Create list of hllBases from a single lbsnRecord"""
        base_records = hll.base_factory(
            facet=facet,
            base=base,
            record=record)
        return base_records

    @classmethod
    def get_hll_metrics(cls, record) -> hll.HllMetrics:
        """Extracts hll metrics based on record type"""
        dict_switcher = {
            lbsn.Origin().DESCRIPTOR.name: cls.get_origin_metrics,
            lbsn.Country().DESCRIPTOR.name: cls.get_country_metrics,
            lbsn.City().DESCRIPTOR.name: cls.get_city_metrics,
            lbsn.Place().DESCRIPTOR.name: cls.get_place_metrics,
            lbsn.User().DESCRIPTOR.name: cls.get_user_metrics,
            lbsn.UserGroup().DESCRIPTOR.name: cls.get_usergroup_metrics,
            lbsn.Post().DESCRIPTOR.name: cls.get_post_metrics,
            lbsn.PostReaction().DESCRIPTOR.name: cls.get_postreaction_metrics,
            lbsn.Relationship().DESCRIPTOR.name: cls.get_relationship_metrics
        }
        extract_function = dict_switcher.get(record.DESCRIPTOR.name)
        record_hll_metrics = extract_function(record)
        return record_hll_metrics

    @staticmethod
    def get_origin_metrics(record) -> hll.HllMetrics:
        """Get hll metrics from lbsn.Origin record"""
        return

    @staticmethod
    def get_country_metrics(record) -> hll.HllMetrics:
        """Get hll metrics from lbsn.Country record"""
        return

    @staticmethod
    def get_city_metrics(record) -> hll.HllMetrics:
        """Get hll metrics from lbsn.City record"""
        return

    @staticmethod
    def get_place_metrics(record) -> hll.HllMetrics:
        """Get hll metrics from lbsn.Place record"""
        place_hll = HLF.hll_concat_origin_guid(record)
        hll_metrics = hll.HllMetrics(place_hll=place_hll)
        return hll_metrics

    @staticmethod
    def get_user_metrics(record) -> hll.HllMetrics:
        """Get hll metrics from lbsn.User record"""
        return

    @staticmethod
    def get_usergroup_metrics(record) -> hll.HllMetrics:
        """Get hll metrics from lbsn.UserGroup record"""
        return

    @staticmethod
    def get_post_metrics(record) -> hll.HllMetrics:
        """Get hll metrics from lbsn.Post record"""
        post_hll = HLF.hll_concat_origin_guid(record)
        user_hll = HLF.hll_concat_user(record)
        date_hll = HLF.hll_concat_userday(record)
        latlng_hll = HLF.hll_concat_latlng(record)
        place_hll = HLF.hll_concat_place(record)
        upt_hll = HLF.hll_concat_upt_hll(record)
        hll_metrics = hll.HllMetrics(
            post_hll=post_hll, user_hll=user_hll,
            date_hll=date_hll, latlng_hll=latlng_hll,
            upt_hll=upt_hll, place_hll=place_hll)
        return hll_metrics

    @staticmethod
    def get_postreaction_metrics(record) -> hll.HllMetrics:
        """Get hll metrics from lbsn.PostReaction record"""
        return

    @staticmethod
    def get_relationship_metrics(record) -> hll.HllMetrics:
        """Get hll metrics from lbsn.Relationship record"""
        return



# -*- coding: utf-8 -*-

"""
Sample bases and metrics for HLL aggregation


The base classes shown here are only examples that illustrate how to extract
typical bases and metrics from Location Based Social Media (LBSN). The
structure is motivated by the 4 facets of Social Media discussed in [1].
Base classes are organized per facet in subfolders base/facet

A template-class is provided that allows extending this module for individual
needs (see class topical.TemplateBase). The general idea is that each class has

* a NAME, which is a reference of type Tuple(str, str) consisting of (1)
    its facet (either spatial, temporal, topical or social) and (2) a
    unique base reference (e.g. latlng, place, region, city or country). In
    our examples, base classes per facet follow the granularity hierarchy
    proposed in [4]
* a Key, which is a unique reference (i.e. the "base") that is measured
* a list of addititional (optional) attributes for the base, e.g. the lat-lng
    key as Postgis Geometry, or a name for the place etc.
* a list of (hll) metrics that are measured for the base (the "overlay"), e.g.
    a list of post_guids ("post_hll"), or user_guids ("user_hll"), or more
    complex metrics such as date_hll (user days, as termed by [2]). These
    lists will be transformed into a "hll shard".

Additional Notes:

* the structure here is closely aligned with the SQL hll files
    maintained in [3], if you add any classes or metrics, make youre they're
    also updated in your hll db
* order is important: the sql commands are constructed dynamically from the
    class structures defined here, thus it is important that order of keys,
    attrs and metrics exactly match the sql db definitions. For this reason,
    OrderedDicts() are used, which store the order in which keys are added
* most classes here make use of (multiple) class inheritance to reduce code
    duplication
* how bases and metrics are mapped from lbsnstructure entirely depends on
    individual needs, the mappings demonstrated here are mere examples. Any
    complex mapping can be added to any of the classes and from any of the
    lbsnstructure objects (e.g. lbsn.Post, lbsn.Place, lbsn.Reaction,
    lbsn.User etc.)

[1] Dunkel, A., Andrienko, G., Andrienko, N., Burghardt, D., Hauthal,
E., & Purves, R. (2019). A conceptual framework for studying collective
reactions to events in location-based social media. International
Journal of Geographical Information Science, 33, 4, 780-804.

[2] Sessions, C., Wood, S. A., Rabotyagov, S., & Fisher, D. M. (2016).
Measuring recreational visitation at U.S. National Parks with
crowd-sourced photo-graphs. Journal of Environmental Management,
183, 703–711. DOI: 10.1016/j.jenvman.2016.09.018

[3] https://gitlab.vgiscience.de/lbsn/databases/hlldb

[4] Löchner, M., Dunkel, A., & Burghardt, D. (2018).
A privacy-aware model to process data from location-based social media.
"""

import inspect
import sys
from collections import OrderedDict, namedtuple
from typing import List

from lbsnstructure import lbsnstructure_pb2 as lbsn
from lbsntransform.tools.helper_functions import HelperFunctions as HF

# named tuple of defined hll metrics
HllMetrics = namedtuple(  # pylint: disable=C0103
    'HllMetricsTuple',
    'user_hll post_hll date_hll latlng_hll upl_hll utl_hll '
    'upt_hll term_hll place_hll', defaults=(None,) * 9)

HllBaseRef = namedtuple('HllBaseRefTuple', 'facet base')
BaseRecordValue = namedtuple('BaseRecordValue', 'record metrics')

# global static variable (dict) to register
# the base classes defined in this module, e.g.
# BASE_REGISTER = {
#     LatLngBase.NAME: LatLngBase,
#     PlaceBase.NAME: PlaceBase,
#     CityBase.NAME: CityBase,
#     ...
# }
BASE_REGISTER = {}

# global static dict for (hard-coded) base header,
# list of sql column names in correct order
# BASE_HEADER = {}
BASE_KEY = {}
BASE_ATTRS = {}
BASE_METRICS = {}


def register_classes():
    """Function to dynamically register base classes for each facet"""
    for facet in ["social", "spatial", "topical", "temporal"]:
        for __, obj in inspect.getmembers(
                sys.modules[f"lbsntransform.output.hll.base.{facet}"]):
            if inspect.isclass(obj):
                if hasattr(obj, 'NAME'):
                    BASE_REGISTER[obj.NAME] = obj
                    BASE_KEY[obj.NAME] = obj().get_key()
                    BASE_ATTRS[obj.NAME] = obj().get_attr_keys()
                    BASE_METRICS[obj.NAME] = obj().get_metric_keys()


def merge_base_metrics(base1, base2):
    """Merge two base-metrics by union of its set values"""
    # merge metric dicts
    for key in base1.metrics.keys():
        new_set = base2.metrics.get(key)
        if new_set is None:
            continue
        base1.metrics[key] |= new_set


def base_factory(facet=None, base=None, record: lbsn.Post = None):
    """Base is initialized based on facet-base tuple
    and constructed by parsing lbsn records

    Any bases that require special hooks need to be registered here. This
    applies, for example, for bases that can appear multiple times
    in a single record (hashtags, emoji, terms etc.).
    """
    records = []
    base_structure = BASE_REGISTER.get((facet, base))
    if base_structure is None:
        return
    # for topical bases (e.g. hashtag, emoji, term)
    # multiple bases can be created
    # from a single lbsn record
    if base == 'hashtag':
        # only explicit hashtags
        tag_terms = HF.filter_terms(record.hashtags)
        for tag in tag_terms:
            records.append(
                base_structure(tag))
    elif base == 'emoji':
        # do nothing
        all_post_emoji = HF.get_all_post_emoji(record.post_body)
        for emoji in all_post_emoji:
            # create base for each term
            records.append(
                base_structure(emoji))
    elif base == 'term':
        # any term mentioned in title,
        # body or hashtag
        all_post_terms = HF.get_all_post_terms(record)
        for term in all_post_terms:
            # create base for each term
            records.append(
                base_structure(term))
    elif base == 'topic':
        raise NotImplementedError(
            "Parsing of Topics base is currently not implemented")
    elif base == 'domain':
        raise NotImplementedError(
            "Parsing of Domains base is currently not implemented")
    elif base == '_term_latlng':
        # any term mentioned in title,
        # body or hashtag
        all_post_terms = HF.get_all_post_terms(record)
        for term in all_post_terms:
            # create base for each term
            records.append(
                base_structure(record=record, term=term))
    elif base == '_hashtag_latlng':
        # any hashtag explicitly used
        tag_terms = HF.filter_terms(record.hashtags)
        for tag in tag_terms:
            records.append(
                base_structure(record=record, hashtag=tag))
    elif base == '_emoji_latlng':
        # any term mentioned in title,
        # body or hashtag
        all_post_emoji = HF.get_all_post_emoji(record.post_body)
        for emoji in all_post_emoji:
            # create base for each term
            records.append(
                base_structure(record=record, emoji=emoji))
    elif base == '_month_hashtag':
        # any hashtag explicitly used
        tag_terms = HF.filter_terms(record.hashtags)
        for tag in tag_terms:
            records.append(
                base_structure(record=record, hashtag=tag))
    else:
        # init for all other bases with single lbsn record
        record = base_structure(record)
        records.append(record)
    return records


class HllBase():
    """Shared attributes for all hll bases"""

    def __init__(self):
        # the key, used as primary (unique) key
        # in the relational db, e.g. 2019-01-01
        # keys can consist of multiple parts, e.g.
        # (lat, lng) which then form composite
        # primary keys in the db
        self.key = OrderedDict()
        # any additional attributes that
        # are stored additionally to the
        # key, e.g. for date-key
        # name: "New Year's Day"
        self.attrs = OrderedDict()
        # the hll-metrics, in constistent
        # order matching the SQL upsert order
        self.metrics = OrderedDict([
            ('user_hll', set()),
            ('post_hll', set())
        ])

    def __ior__(self, other):
        """Implements bitwise or using the | operator.
        """
        if other is None:
            return self
        merge_base_metrics(self, other)
        return self

    def __repr__(self):
        """Implement custom format str for debug"""
        return HF.format_base_repr(self)

    def get_key_value(self):
        """Returns key value for base
        """
        return tuple(self.key.values())

    def get_key(self):
        """Returns key name for base
        """
        return list(self.key.keys())

    def get_attr_keys(self):
        """Returns attr keys for base
        """
        return list(self.attrs.keys())

    def get_metric_keys(self):
        """Returns metric keys for base
        """
        return list(self.metrics.keys())

    def get_sql_header(self) -> List[str]:
        """Get joined header for hll upsert sql
        Concat column names for key, attrs and metrics, e.g.:
        latitude, longitude, latlng_geom, user_hll, post_hll, date_hll, utl_hll
        """
        base_key_cols = [self.key.keys()]
        base_attr_cols = [self.attrs.keys()]
        base_metrics_cols = [self.metrics.keys()]
        return base_key_cols + base_attr_cols + base_metrics_cols

    def get_prepared_record(self):
        """Return prepared sql values tuple

        Consisting of
        * key and attributes tuple (base_record) = the record
        * metric dicts with values = the metrics
        """
        base_record = tuple(self.key.values())
        for attr in self.attrs.values():
            base_record += (attr,)
        return BaseRecordValue(base_record, self.metrics)
